{ **********************************************************************
  *                          Unit MATRICES.PAS                         *
  *                             Version 2.4                            *
  *                      (c) J. Debord, June 2004                      *
  **********************************************************************
  This unit implements dynamic allocation of vectors and matrices in
  Pascal, together with various matrix operations.

  Dynamic allocation is allowed by declaring arrays as pointers. There
  are 8 types available :

  PVector,     PMatrix      for floating point arrays
  PIntVector,  PIntMatrix   for integer arrays
  PBoolVector, PBoolMatrix  for boolean arrays
  PStrVector,  PStrMatrix   for string arrays (255 char.)

  To use these arrays in your programs, you must :

  (1) Declare variables of the appropriate type, e.g.

         var
           V : PVector;
           A : PMatrix;

  (2) Allocate each array BEFORE using it :

         DimVector(V, N);         creates vector V[0..N]
         DimMatrix(A, N, M);      creates matrix A[0..N, 0..M]
                                  where N, M are two integer variables

         If the allocation succeeds, all array elements are initialized
         to zero (for numeric arrays), False (for boolean arrays), or
         the null string (for string arrays). Otherwise, the pointer is
         initialized to NIL.

  (3) Use arrays as in standard Turbo Pascal, with the following
      exceptions :

      (a) You must use the indirection operator (^) to reference any
          array element, i.e. write V^[I] and A^[I]^[J] instead of
          V[I] and A[I,J].

      (b) You cannot use the assignment operator (:=) to copy the
          contents of an array into another array. Writing B := A
          simply makes B point to the same memory block than A. You
          must use one of the provided Copy... procedures (see their
          documentation in the interface part of the unit).

      In addition, note that :

      (a) All arrays begin at index 0, so that the 0-indexed element
          is always present, even if you don't use it.

      (b) A matrix is declared as an array of vectors, so that A^[I]
          denotes the I-th vector of matrix A and may be used as any
          vector.

  (4) Deallocate arrays when you no longer need them. This will free
      the corresponding memory :

         DelVector(V, N);
         DelMatrix(A, N, M);

  For more information, read the comments of each routine in the
  interface part of the unit, and check the demo programs.
  **********************************************************************
  References :
  1) 'Mathematiques et Statistiques' by H. Haut (PSI ed.) : GaussJordan
  2) EISPACK (http://www.netlib.org/eispack) : SV_Decomp
  3) 'Matrix Computations' by Golub & Van Loan : QR_Decomp & QR_Solve
     (Pascal implementation contributed by Mark Vaughan)
  ********************************************************************** }

unit Matrices;

interface

uses
  FMath;

{ **********************************************************************
  This section defines some error codes
  ********************************************************************** }

const
  MAT_OK       =   0;  { No error }
  MAT_SINGUL   = - 1;  { Singular matrix }
  MAT_NON_CONV = - 2;  { Non convergence of iterative procedure }
  MAT_NOT_PD   = - 3;  { Matrix not positive definite }

{ **********************************************************************
  This section defines the vector and matrix types
  ********************************************************************** }

const
{$IFDEF EXTENDEDREAL}
  MAX_FLT = 6552;      { Max size of real vector }
{$ELSE}
{$IFDEF SINGLEREAL}
  MAX_FLT = 16382;
{$ELSE}
{$IFDEF PASCALREAL}
  MAX_FLT = 10921;
{$ELSE}
  {$DEFINE DOUBLEREAL}
  MAX_FLT = 8190;
{$ENDIF}
{$ENDIF}
{$ENDIF}

  MAX_INT  = 16382;  { Max size of integer vector }
  MAX_BOOL = 32766;  { Max size of boolean vector }
  MAX_STR  = 254;    { Max size of string vector }
  MAX_VEC  = 16382;  { Max number of vectors in a matrix }

type
  TVector     = array[0..MAX_FLT] of Float;
  TIntVector  = array[0..MAX_INT] of Integer;
  TBoolVector = array[0..MAX_BOOL] of Boolean;
  TStrVector  = array[0..MAX_STR] of String;

  PVector     = ^TVector;
  PIntVector  = ^TIntVector;
  PBoolVector = ^TBoolVector;
  PStrVector  = ^TStrVector;

  TMatrix     = array[0..MAX_VEC] of PVector;
  TIntMatrix  = array[0..MAX_VEC] of PIntVector;
  TBoolMatrix = array[0..MAX_VEC] of PBoolVector;
  TStrMatrix  = array[0..MAX_VEC] of PStrVector;

  PMatrix     = ^TMatrix;
  PIntMatrix  = ^TIntMatrix;
  PBoolMatrix = ^TBoolMatrix;
  PStrMatrix  = ^TStrMatrix;

{ **********************************************************************
  Memory allocation routines
  ********************************************************************** }

procedure DimVector(var V : PVector; Ubound : Integer);
{ ----------------------------------------------------------------------
  Creates floating point vector V[0..Ubound]
  ---------------------------------------------------------------------- }

procedure DimIntVector(var V : PIntVector; Ubound : Integer);
{ ----------------------------------------------------------------------
  Creates integer vector V[0..Ubound]
  ---------------------------------------------------------------------- }

procedure DimBoolVector(var V : PBoolVector; Ubound : Integer);
{ ----------------------------------------------------------------------
  Creates boolean vector V[0..Ubound]
  ---------------------------------------------------------------------- }

procedure DimStrVector(var V : PStrVector; Ubound : Integer);
{ ----------------------------------------------------------------------
  Creates string vector V[0..Ubound]
  ---------------------------------------------------------------------- }

procedure DimMatrix(var A : PMatrix; Ubound1, Ubound2 : Integer);
{ ----------------------------------------------------------------------
  Creates floating point matrix A[0..Ubound1, 0..Ubound2]
  ---------------------------------------------------------------------- }

procedure DimIntMatrix(var A : PIntMatrix; Ubound1, Ubound2 : Integer);
{ ----------------------------------------------------------------------
  Creates integer matrix A[0..Ubound1, 0..Ubound2]
  ---------------------------------------------------------------------- }

procedure DimBoolMatrix(var A : PBoolMatrix; Ubound1, Ubound2 : Integer);
{ ----------------------------------------------------------------------
  Creates boolean matrix A[0..Ubound1, 0..Ubound2]
  ---------------------------------------------------------------------- }

procedure DimStrMatrix(var A : PStrMatrix; Ubound1, Ubound2 : Integer);
{ ----------------------------------------------------------------------
  Creates string matrix A[0..Ubound1, 0..Ubound2]
  ---------------------------------------------------------------------- }

{ **********************************************************************
  Memory deallocation routines
  ********************************************************************** }

procedure DelVector(var V : PVector; Ubound : Integer);
{ ----------------------------------------------------------------------
  Deletes floating point vector V[0..Ubound]
  ---------------------------------------------------------------------- }

procedure DelIntVector(var V : PIntVector; Ubound : Integer);
{ ----------------------------------------------------------------------
  Deletes integer vector V[0..Ubound]
  ---------------------------------------------------------------------- }

procedure DelBoolVector(var V : PBoolVector; Ubound : Integer);
{ ----------------------------------------------------------------------
  Deletes boolean vector V[0..Ubound]
  ---------------------------------------------------------------------- }

procedure DelStrVector(var V : PStrVector; Ubound : Integer);
{ ----------------------------------------------------------------------
  Deletes string vector V[0..Ubound]
  ---------------------------------------------------------------------- }

procedure DelMatrix(var A : PMatrix; Ubound1, Ubound2 : Integer);
{ ----------------------------------------------------------------------
  Deletes floating point matrix A[0..Ubound1, 0..Ubound2]
  ---------------------------------------------------------------------- }

procedure DelIntMatrix(var A : PIntMatrix; Ubound1, Ubound2 : Integer);
{ ----------------------------------------------------------------------
  Deletes integer matrix A[0..Ubound1, 0..Ubound2]
  ---------------------------------------------------------------------- }

procedure DelBoolMatrix(var A : PBoolMatrix; Ubound1, Ubound2 : Integer);
{ ----------------------------------------------------------------------
  Deletes boolean matrix A[0..Ubound1, 0..Ubound2]
  ---------------------------------------------------------------------- }

procedure DelStrMatrix(var A : PStrMatrix; Ubound1, Ubound2 : Integer);
{ ----------------------------------------------------------------------
  Deletes string matrix A[0..Ubound1, 0..Ubound2]
  ---------------------------------------------------------------------- }

{ **********************************************************************
  Routines for copying vectors and matrices
  ----------------------------------------------------------------------
  Lbound, Ubound   : indices of first and last vector elements
  Lbound1, Lbound2 : indices of first matrix element in each dimension
  Ubound1, Ubound2 : indices of last matrix element in each dimension
  ********************************************************************** }

procedure SwapRows(I, K : Integer; A : PMatrix; Lbound, Ubound : Integer);
{ ----------------------------------------------------------------------
  Exchanges rows I and K of matrix A
  ---------------------------------------------------------------------- }

procedure SwapCols(J, K : Integer; A : PMatrix; Lbound, Ubound : Integer);
{ ----------------------------------------------------------------------
  Exchanges columns J and K of matrix A
  ---------------------------------------------------------------------- }

procedure CopyVector(Dest, Source : PVector; Lbound, Ubound : Integer);
{ ----------------------------------------------------------------------
  Copies vector Source into vector Dest
  ---------------------------------------------------------------------- }

procedure CopyMatrix(Dest, Source : PMatrix;
                     Lbound1, Lbound2, Ubound1, Ubound2 : Integer);
{ ----------------------------------------------------------------------
  Copies matrix Source into matrix Dest
  ---------------------------------------------------------------------- }

procedure CopyRowFromVector(Dest : PMatrix; Source : PVector;
                            Lbound, Ubound, Row : Integer);
{ ----------------------------------------------------------------------
  Copies vector Source into line Row of matrix Dest
  ---------------------------------------------------------------------- }

procedure CopyColFromVector(Dest : PMatrix; Source : PVector;
                            Lbound, Ubound, Col : Integer);
{ ----------------------------------------------------------------------
  Copies vector Source into column Col of matrix Dest
  ---------------------------------------------------------------------- }

procedure CopyVectorFromRow(Dest : PVector; Source : PMatrix;
                            Lbound, Ubound, Row : Integer);
{ ----------------------------------------------------------------------
  Copies line Row of matrix Source into vector Dest
  ---------------------------------------------------------------------- }

procedure CopyVectorFromCol(Dest : PVector; Source : PMatrix;
                            Lbound, Ubound, Col : Integer);
{ ----------------------------------------------------------------------
  Copies column Col of matrix Source into vector Dest
  ---------------------------------------------------------------------- }

{ **********************************************************************
  Vector and matrix functions
  ********************************************************************** }

function Min(X : PVector; Lbound, Ubound : Integer) : Float;
{ ----------------------------------------------------------------------
  Returns the lowest value of vector X
  ---------------------------------------------------------------------- }

function Max(X : PVector; Lbound, Ubound : Integer) : Float;
{ ----------------------------------------------------------------------
  Returns the highest value of vector X
  ---------------------------------------------------------------------- }

function IntMin(X : PIntVector; Lbound, Ubound : Integer) : Integer;
{ ----------------------------------------------------------------------
  Returns the lowest value of integer vector X
  ---------------------------------------------------------------------- }

function IntMax(X : PIntVector; Lbound, Ubound : Integer) : Integer;
{ ----------------------------------------------------------------------
  Returns the highest value of integer vector X
  ---------------------------------------------------------------------- }

procedure Transpose(A : PMatrix;
                    Lbound1, Lbound2, Ubound1, Ubound2 : Integer;
                    A_t : PMatrix);
{ ----------------------------------------------------------------------
  Transposes a matrix
  ----------------------------------------------------------------------
  Input parameters : A       = original matrix
                     Lbound1,
                     Lbound2 = indices of 1st matrix elem. in each dim.
                     Ubound1,
                     Ubound2 = indices of last matrix elem. in each dim.
  ----------------------------------------------------------------------
  Output parameter : A_t     = transposed matrix
  ---------------------------------------------------------------------- }

function GaussJordan(A : PMatrix; Lbound, Ubound1, Ubound2 : Integer;
                     var Det : Float) : Integer;
{ ----------------------------------------------------------------------
  Transforms a matrix according to the Gauss-Jordan method. The original
  matrix A is a (n x m) matrix, with n <= m. The first n columns of A
  correspond to a matrix B, and the remaining columns correspond to a
  matrix C. The transformed matrix is such that its first n columns
  contain the inverse matrix B^(-1) and its remaining columns contain
  the solution matrix X, such that B * X = C. The determinant of B is
  returned in Det. If n = m, only the inverse matrix and determinant
  are returned.
  ----------------------------------------------------------------------
  Input parameters  : A       = original matrix
                      Lbound  = index of first matrix element
                      Ubound1 = index of last matrix element in 1st dim.
                      Ubound2 = index of last matrix element in 2nd dim.
  ----------------------------------------------------------------------
  Output parameters : A   = transformed matrix
                      Det = determinant
  ----------------------------------------------------------------------
  Possible results  : MAT_OK
                      MAT_SINGUL
  ----------------------------------------------------------------------
  NB : This procedure destroys the original matrix A
  ---------------------------------------------------------------------- }

function Cholesky(A : PMatrix; Lbound, Ubound : Integer;
                  L : PMatrix) : Integer;
{ ----------------------------------------------------------------------
  Cholesky decomposition. Factors the symmetric positive definite matrix
  A as a product L * L', where L is a lower triangular matrix. This
  procedure may be used as a test of positive definiteness.
  ----------------------------------------------------------------------
  Input parameters : A      = matrix
                     Lbound = index of first matrix element
                     Ubound = index of last matrix element
  ----------------------------------------------------------------------
  Output parameter : L      = Cholesky factor of matrix A
  ----------------------------------------------------------------------
  Possible results : MAT_OK
                     MAT_NOT_PD
  ---------------------------------------------------------------------- }

function LU_Decomp(A : PMatrix; Lbound, Ubound : Integer) : Integer;
{ ----------------------------------------------------------------------
  LU decomposition. Factors the square matrix A as a product L * U,
  where L is a lower triangular matrix (with unit diagonal terms) and U
  is an upper triangular matrix. This routine is used in conjunction
  with LU_Solve to solve a system of equations.
  ----------------------------------------------------------------------
  Input parameters : A      = matrix
                     Lbound = index of first matrix element
                     Ubound = index of last matrix element
  ----------------------------------------------------------------------
  Output parameter : A      = contains the elements of L and U
  ----------------------------------------------------------------------
  Possible results : MAT_OK
                     MAT_SINGUL
  ----------------------------------------------------------------------
  NB : This procedure destroys the original matrix A
  ---------------------------------------------------------------------- }

procedure LU_Solve(A : PMatrix; B : PVector; Lbound, Ubound : Integer;
                   X : PVector);
{ ----------------------------------------------------------------------
  Solves a system of equations whose matrix has been transformed by
  LU_Decomp
  ----------------------------------------------------------------------
  Input parameters : A      = result from LU_Decomp
                     B      = constant vector
                     Lbound,
                     Ubound = as in LU_Decomp
  ----------------------------------------------------------------------
  Output parameter : X      = solution vector
  ---------------------------------------------------------------------- }

function SV_Decomp(A : PMatrix; Lbound, Ubound1, Ubound2 : Integer;
                   S : PVector; V : PMatrix) : Integer;
{ ----------------------------------------------------------------------
  Singular value decomposition. Factors the matrix A (n x m, with n >= m)
  as a product U * S * V' where U is a (n x m) column-orthogonal matrix,
  S a (m x m) diagonal matrix with elements >= 0 (the singular values)
  and V a (m x m) orthogonal matrix. This routine is used in conjunction
  with SV_Solve to solve a system of equations.
  ----------------------------------------------------------------------
  Input parameters : A       = matrix
                     Lbound  = index of first matrix element
                     Ubound1 = index of last matrix element in 1st dim.
                     Ubound2 = index of last matrix element in 2nd dim.
  ----------------------------------------------------------------------
  Output parameter : A       = contains the elements of U
                     S       = vector of singular values
                     V       = orthogonal matrix
  ----------------------------------------------------------------------
  Possible results  :  0 : No error
                      -i : Non-convergence has been encountered during
                           the search for the i-th singular value. The
                           singular values should be correct for indices
                           (i+1)..Ubound, as well as the corresponding
                           columns of U and V.
  ----------------------------------------------------------------------
  NB : This procedure destroys the original matrix A
  ---------------------------------------------------------------------- }

procedure SV_SetZero(S : PVector; Lbound, Ubound : Integer; Tol : Float);
{ ----------------------------------------------------------------------
  Sets the singular values to zero if they are lower than a specified
  threshold.
  ----------------------------------------------------------------------
  Input parameters : S      = vector of singular values
                     Tol    = relative tolerance
                              Threshold value will be Tol * Max(S)
                     Lbound = index of first vector element
                     Ubound = index of last vector element
  ----------------------------------------------------------------------
  Output parameter : S      = modified singular values
  ---------------------------------------------------------------------- }

procedure SV_Solve(U : PMatrix; S : PVector; V : PMatrix; B : PVector;
                   Lbound, Ubound1, Ubound2 : Integer;
                   X : PVector);
{ ----------------------------------------------------------------------
  Solves a system of equations by singular value decomposition, after
  the matrix has been transformed by SV_Decomp, and the lowest singular
  values have been set to zero by SV_SetZero.
  ----------------------------------------------------------------------
  Input parameters : U, S, V = vector and matrices from SV_Decomp
                     B       = constant vector
                     Lbound,
                     Ubound1,
                     Ubound2 = as in SV_Decomp
  ----------------------------------------------------------------------
  Output parameter : X       = solution vector
                             = V * Diag(1/s(i)) * U' * B, for s(i) <> 0
  ---------------------------------------------------------------------- }

procedure SV_Approx(U : PMatrix; S : PVector; V : PMatrix;
                    Lbound, Ubound1, Ubound2 : Integer;
                    A : PMatrix);
{ ----------------------------------------------------------------------
  Approximates a matrix A by the product USV', after the lowest singular
  values have been set to zero by SV_SetZero.
  ----------------------------------------------------------------------
  Input parameters : U, S, V = vector and matrices from SV_Decomp
                     Lbound,
                     Ubound1,
                     Ubound2 = as in SV_Decomp
  ----------------------------------------------------------------------
  Output parameter : A       = approximated matrix
  ---------------------------------------------------------------------- }

function QR_Decomp(A : PMatrix; Lbound, Ubound1, Ubound2 : Integer;
                   R : PMatrix) : Integer;
{ ----------------------------------------------------------------------
  QR decomposition. Factors the matrix A (n x m, with n >= m) as a
  product Q * R where Q is a (n x m) column-orthogonal matrix, and R
  a (m x m) upper triangular matrix. This routine is used in conjunction
  with QR_Solve to solve a system of equations.
  ----------------------------------------------------------------------
  Input parameters : A       = matrix
                     Lbound  = index of first matrix element
                     Ubound1 = index of last matrix element in 1st dim.
                     Ubound2 = index of last matrix element in 2nd dim.
  ----------------------------------------------------------------------
  Output parameter : A       = contains the elements of Q
                     R       = upper triangular matrix
  ----------------------------------------------------------------------
  Possible results : MAT_OK
                     MAT_SING
  ----------------------------------------------------------------------
  NB : This procedure destroys the original matrix A
  ---------------------------------------------------------------------- }

procedure QR_Solve(Q, R : PMatrix; B : PVector;
                   Lbound, Ubound1, Ubound2 : Integer;
                   X : PVector);
{ ----------------------------------------------------------------------
  Solves a system of equations by the QR decomposition,
  after the matrix has been transformed by QR_Decomp.
  ----------------------------------------------------------------------
  Input parameters : Q, R    = matrices from QR_Decomp
                     B       = constant vector
                     Lbound,
                     Ubound1,
                     Ubound2 = as in QR_Decomp
  ----------------------------------------------------------------------
  Output parameter : X       = solution vector
  ---------------------------------------------------------------------- }

implementation

  procedure DimVector(var V : PVector; Ubound : Integer);
  var
    I : Integer;
  begin
    { Check bounds }
    if (Ubound < 0) or (Ubound > MAX_FLT) then
      begin
        V := nil;
        Exit;
      end;

    { Allocate vector }
    GetMem(V, Succ(Ubound) * SizeOf(Float));
    if V = nil then Exit;

    { Initialize vector }
    for I := 0 to Ubound do
      V^[I] := 0.0;
  end;

  procedure DimIntVector(var V : PIntVector; Ubound : Integer);
  var
    I : Integer;
  begin
    { Check bounds }
    if (Ubound < 0) or (Ubound > MAX_INT) then
      begin
        V := nil;
        Exit;
      end;

    { Allocate vector }
    GetMem(V, Succ(Ubound) * SizeOf(Integer));
    if V = nil then Exit;

    { Initialize vector }
    for I := 0 to Ubound do
      V^[I] := 0;
  end;

  procedure DimBoolVector(var V : PBoolVector; Ubound : Integer);
  var
    I : Integer;
  begin
    { Check bounds }
    if (Ubound < 0) or (Ubound > MAX_BOOL) then
      begin
        V := nil;
        Exit;
      end;

    { Allocate vector }
    GetMem(V, Succ(Ubound) * SizeOf(Boolean));
    if V = nil then Exit;

    { Initialize vector }
    for I := 0 to Ubound do
      V^[I] := False;
  end;

  procedure DimStrVector(var V : PStrVector; Ubound : Integer);
  var
    I : Integer;
  begin
    { Check bounds }
    if (Ubound < 0) or (Ubound > MAX_STR) then
      begin
        V := nil;
        Exit;
      end;

    { Allocate vector }
    GetMem(V, Succ(Ubound) * 256);
    if V = nil then Exit;

    { Initialize vector }
    for I := 0 to Ubound do
      V^[I] := '';
  end;

  procedure DimMatrix(var A : PMatrix; Ubound1, Ubound2 : Integer);
  var
    I, J : Integer;
    RowSize : Word;
  begin
    if (Ubound1 < 0) or (Ubound1 > MAX_VEC) or
    (Ubound2 < 0) or (Ubound2 > MAX_FLT) then
      begin
        A := nil;
        Exit;
      end;

    { Allocate matrix }
    GetMem(A, Succ(Ubound1) * SizeOf(PVector));
    if A = nil then Exit;

    { Size of a row }
    RowSize := Succ(Ubound2) * SizeOf(Float);

    { Allocate each row }
    for I := 0 to Ubound1 do
      begin
        GetMem(A^[I], RowSize);
        if A^[I] = nil then
          begin
            A := nil;
            Exit;
          end;
      end;

    { Initialize matrix }
    for I := 0 to Ubound1 do
      for J := 0 to Ubound2 do
        A^[I]^[J] := 0.0;
  end;

  procedure DimIntMatrix(var A : PIntMatrix; Ubound1, Ubound2 : Integer);
  var
    I, J : Integer;
    RowSize : Word;
  begin
    { Check bounds }
    if (Ubound1 < 0) or (Ubound1 > MAX_VEC) or
    (Ubound2 < 0) or (Ubound2 > MAX_INT) then
      begin
        A := nil;
        Exit;
      end;

    { Allocate matrix }
    GetMem(A, Succ(Ubound1) * SizeOf(PIntVector));
    if A = nil then Exit;

    { Size of a row }
    RowSize := Succ(Ubound2) * SizeOf(Integer);

    { Allocate each row }
    for I := 0 to Ubound1 do
      begin
        GetMem(A^[I], RowSize);
        if A^[I] = nil then
          begin
            A := nil;
            Exit;
          end;
      end;

    { Initialize matrix }
    for I := 0 to Ubound1 do
      for J := 0 to Ubound2 do
        A^[I]^[J] := 0;
  end;

  procedure DimBoolMatrix(var A : PBoolMatrix; Ubound1, Ubound2 : Integer);
  var
    I, J : Integer;
    RowSize : Word;
  begin
    { Check bounds }
    if (Ubound1 < 0) or (Ubound1 > MAX_VEC) or
    (Ubound2 < 0) or (Ubound2 > MAX_BOOL) then
      begin
        A := nil;
        Exit;
      end;

    { Allocate matrix }
    GetMem(A, Succ(Ubound1) * SizeOf(PBoolVector));
    if A = nil then Exit;

    { Size of a row }
    RowSize := Succ(Ubound2) * SizeOf(Boolean);

    { Allocate each row }
    for I := 0 to Ubound1 do
      begin
        GetMem(A^[I], RowSize);
        if A^[I] = nil then
          begin
            A := nil;
            Exit;
          end;
      end;

    { Initialize matrix }
    for I := 0 to Ubound1 do
      for J := 0 to Ubound2 do
        A^[I]^[J] := False;
  end;

  procedure DimStrMatrix(var A : PStrMatrix; Ubound1, Ubound2 : Integer);
  var
    I, J : Integer;
    RowSize : Word;
  begin
    { Check bounds }
    if (Ubound1 < 0) or (Ubound1 > MAX_VEC) or
    (Ubound2 < 0) or (Ubound2 > MAX_STR) then
      begin
        A := nil;
        Exit;
      end;

    { Allocate matrix }
    GetMem(A, Succ(Ubound1) * SizeOf(PStrVector));
    if A = nil then Exit;

    { Size of a row }
    RowSize := Succ(Ubound2) * 256;

    { Allocate each row }
    for I := 0 to Ubound1 do
      begin
        GetMem(A^[I], RowSize);
        if A^[I] = nil then
          begin
            A := nil;
            Exit;
          end;
      end;

    { Initialize matrix }
    for I := 0 to Ubound1 do
      for J := 0 to Ubound2 do
        A^[I]^[J] := '';
  end;

  procedure DelVector(var V : PVector; Ubound : Integer);
  begin
    if V <> nil then
      begin
        FreeMem(V, Succ(Ubound) * SizeOf(Float));
        V := nil;
      end;
  end;

  procedure DelIntVector(var V : PIntVector; Ubound : Integer);
  begin
    if V <> nil then
      begin
        FreeMem(V, Succ(Ubound) * SizeOf(Integer));
        V := nil;
      end;
  end;

  procedure DelBoolVector(var V : PBoolVector; Ubound : Integer);
  begin
    if V <> nil then
      begin
        FreeMem(V, Succ(Ubound) * SizeOf(Boolean));
        V := nil;
      end;
  end;

  procedure DelStrVector(var V : PStrVector; Ubound : Integer);
  begin
    if V <> nil then
      begin
        FreeMem(V, Succ(Ubound) * 256);
        V := nil;
      end;
  end;

  procedure DelMatrix(var A : PMatrix; Ubound1, Ubound2 : Integer);
  var
    I : Integer;
    RowSize : Word;
  begin
    if A <> nil then
      begin
        RowSize := Succ(Ubound2) * SizeOf(Float);
        for I := Ubound1 downto 0 do
          FreeMem(A^[I], RowSize);
        FreeMem(A, Succ(Ubound1) * SizeOf(PVector));
        A := nil;
      end;
  end;

  procedure DelIntMatrix(var A : PIntMatrix; Ubound1, Ubound2 : Integer);
  var
    I : Integer;
    RowSize : Word;
  begin
    if A <> nil then
      begin
        RowSize := Succ(Ubound2) * SizeOf(Integer);
        for I := Ubound1 downto 0 do
          FreeMem(A^[I], RowSize);
        FreeMem(A, Succ(Ubound1) * SizeOf(PIntVector));
        A := nil;
      end;
  end;

  procedure DelBoolMatrix(var A : PBoolMatrix; Ubound1, Ubound2 : Integer);
  var
    I : Integer;
    RowSize : Word;
  begin
    if A <> nil then
      begin
        RowSize := Succ(Ubound2) * SizeOf(Boolean);
        for I := Ubound1 downto 0 do
          FreeMem(A^[I], RowSize);
        FreeMem(A, Succ(Ubound1) * SizeOf(PBoolVector));
        A := nil;
      end;
  end;

  procedure DelStrMatrix(var A : PStrMatrix; Ubound1, Ubound2 : Integer);
  var
    I : Integer;
    RowSize : Word;
  begin
    if A <> nil then
      begin
        RowSize := Succ(Ubound2) * 256;
        for I := Ubound1 downto 0 do
          FreeMem(A^[I], RowSize);
        FreeMem(A, Succ(Ubound1) * SizeOf(PStrVector));
        A := nil;
      end;
  end;

  procedure SwapRows(I, K : Integer; A : PMatrix; Lbound, Ubound : Integer);
  var
    J : Integer;
  begin
    for J := Lbound to Ubound do
      FSwap(A^[I]^[J], A^[K]^[J]);
  end;

  procedure SwapCols(J, K : Integer; A : PMatrix; Lbound, Ubound : Integer);
  var
    I : Integer;
  begin
    for I := Lbound to Ubound do
      FSwap(A^[I]^[J], A^[I]^[K]);
  end;

  procedure CopyVector(Dest, Source : PVector; Lbound, Ubound : Integer);
  var
    I : Integer;
  begin
    for I := Lbound to Ubound do
      Dest^[I] := Source^[I];
  end;

  procedure CopyMatrix(Dest, Source : PMatrix;
                       Lbound1, Lbound2, Ubound1, Ubound2 : Integer);
  var
    I, J : Integer;
  begin
    for I := Lbound1 to Ubound1 do
      for J := Lbound2 to Ubound2 do
        Dest^[I]^[J] := Source^[I]^[J];
  end;

  procedure CopyRowFromVector(Dest : PMatrix; Source : PVector;
                              Lbound, Ubound, Row : Integer);
  var
    J : Integer;
  begin
    for J := Lbound to Ubound do
      Dest^[Row]^[J] := Source^[J];
  end;

  procedure CopyColFromVector(Dest : PMatrix; Source : PVector;
                              Lbound, Ubound, Col : Integer);
  var
    I : Integer;
  begin
    for I := Lbound to Ubound do
      Dest^[I]^[Col] := Source^[I];
  end;

  procedure CopyVectorFromRow(Dest : PVector; Source : PMatrix;
                              Lbound, Ubound, Row : Integer);
  var
    J : Integer;
  begin
    for J := Lbound to Ubound do
      Dest^[J] := Source^[Row]^[J];
  end;

  procedure CopyVectorFromCol(Dest : PVector; Source : PMatrix;
                              Lbound, Ubound, Col : Integer);
  var
    I : Integer;
  begin
    for I := Lbound to Ubound do
      Dest^[I] := Source^[I]^[Col];
  end;

  function Min(X : PVector; Lbound, Ubound : Integer) : Float;
  var
    Xmin : Float;
    I : Integer;
  begin
    Xmin := X^[Lbound];
    for I := Succ(Lbound) to Ubound do
      if X^[I] < Xmin then Xmin := X^[I];
    Min := Xmin;
  end;

  function Max(X : PVector; Lbound, Ubound : Integer) : Float;
  var
    Xmax : Float;
    I : Integer;
  begin
    Xmax := X^[Lbound];
    for I := Succ(Lbound) to Ubound do
      if X^[I] > Xmax then Xmax := X^[I];
    Max := Xmax;
  end;

  function IntMin(X : PIntVector; Lbound, Ubound : Integer) : Integer;
  var
    I, Xmin : Integer;
  begin
    Xmin := X^[Lbound];
    for I := Succ(Lbound) to Ubound do
      if X^[I] < Xmin then Xmin := X^[I];
    IntMin := Xmin;
  end;

  function IntMax(X : PIntVector; Lbound, Ubound : Integer) : Integer;
  var
    I, Xmax : Integer;
  begin
    Xmax := X^[Lbound];
    for I := Succ(Lbound) to Ubound do
      if X^[I] > Xmax then Xmax := X^[I];
    IntMax := Xmax;
  end;

  procedure Transpose(A : PMatrix;
                      Lbound1, Lbound2, Ubound1, Ubound2 : Integer;
                      A_t : PMatrix);
  var
    I, J : Integer;
  begin
    for I := Lbound1 to Ubound1 do
      for J := Lbound2 to Ubound2 do
        A_t^[J]^[I] := A^[I]^[J];
  end;

  function GaussJordan(A : PMatrix; Lbound, Ubound1, Ubound2 : Integer;
                       var Det : Float) : Integer;
  var
    Pvt        : Float;       { Pivot }
    Ik, Jk     : Integer;     { Pivot's row and column }
    PRow, PCol : PIntVector;  { Stores pivot's row and column }
    MCol       : PVector;     { Stores a column of matrix A }
    I, J, K    : Integer;     { Loop variables }
    T          : Float;       { Temporary variable }
  begin
    DimIntVector(PRow, Ubound1);
    DimIntVector(PCol, Ubound1);
    DimVector(MCol, Ubound1);

    Det := 1.0;

    K := Lbound;
    while K <= Ubound1 do
      begin
        { Search for largest pivot in submatrix
          A[K..Ubound1, K..Ubound1] }
        Pvt := A^[K]^[K];
        Ik := K;
        Jk := K;
        for I := K to Ubound1 do
          for J := K to Ubound1 do
            if Abs(A^[I]^[J]) > Abs(Pvt) then
              begin
                Pvt := A^[I]^[J];
                Ik := I;
                Jk := J;
              end;

        { Pivot too small ==> quasi-singular matrix }
        if Abs(Pvt) < MACHEP then
          begin
            DelIntVector(PRow, Ubound1);
            DelIntVector(PCol, Ubound1);
            DelVector(MCol, Ubound1);
            GaussJordan := MAT_SINGUL;
            Exit;
          end;

        { Save pivot position }
        PRow^[K] := Ik;
        PCol^[K] := Jk;

        { Update determinant }
        Det := Det * Pvt;
        if Ik <> K then Det := - Det;
        if Jk <> K then Det := - Det;

        { Exchange current row (K) with pivot row (Ik) }
        if Ik <> K then
          SwapRows(Ik, K, A, Lbound, Ubound2);

        { Exchange current column (K) with pivot column (Jk) }
        if Jk <> K then
          SwapCols(Jk, K, A, Lbound, Ubound1);

        { Store col. K of A_inv into MCol and set this col. to 0 }
        for I := Lbound to Ubound1 do
          if I <> K then
            begin
              MCol^[I] := A^[I]^[K];
              A^[I]^[K] := 0.0;
            end
          else
            begin
              MCol^[I] := 0.0;
              A^[I]^[K] := 1.0;
            end;

        { Transform pivot row }
        for J := Lbound to Ubound2 do
          A^[K]^[J] := A^[K]^[J] / Pvt;

        { Transform other rows }
        for I := Lbound to Ubound1 do
          if I <> K then
            for J := Lbound to Ubound2 do
              A^[I]^[J] := A^[I]^[J] - MCol^[I] * A^[K]^[J];

        Inc(K);
      end;

    { Exchange rows of inverse matrix }
    for I := Ubound1 downto Lbound do
      begin
        Ik := PCol^[I];
        if Ik <> I then
          SwapRows(Ik, I, A, Lbound, Ubound2);
      end;

    { Exchange columns of inverse matrix }
    for J := Ubound1 downto Lbound do
      begin
        Jk := PRow^[J];
        SwapCols(Jk, J, A, Lbound, Ubound1);
      end;

    DelIntVector(PRow, Ubound1);
    DelIntVector(PCol, Ubound1);
    DelVector(MCol, Ubound1);

    GaussJordan := MAT_OK;
  end;

  function Cholesky(A : PMatrix; Lbound, Ubound : Integer;
                    L : PMatrix) : Integer;
  var
    I, J, K : Integer;
    Sum : Float;
  begin
    for K := Lbound to Ubound do
      begin
        Sum := A^[K]^[K];
        for J := Lbound to K - 1 do
          Sum := Sum - Sqr(L^[K]^[J]);

        if Sum <= 0.0 then
          begin
            Cholesky := MAT_NOT_PD;
            Exit;
          end;

        L^[K]^[K] := Sqrt(Sum);
        for I := K + 1 to Ubound do
          begin
            Sum := A^[I]^[K];
            for J := Lbound to K - 1 do
              Sum := Sum - L^[I]^[J] * L^[K]^[J];
            L^[I]^[K] := Sum / L^[K]^[K];
          end;
      end;
    Cholesky := MAT_OK;
  end;

var                      { Used by LU procedures }
  LastDim : Integer;     { Dimension of the last system solved }
  Index   : PIntVector;  { Records the row permutations }

  function LU_Decomp(A : PMatrix; Lbound, Ubound : Integer) : Integer;
  const
    TINY = 1.0E-20;
  var
    I, Imax, J, K : Integer;
    Pvt, T, Sum : Float;
    V : PVector;
  begin
    DimVector(V, Ubound);
    { Reallocate Index }
    if Index <> nil then
      DelIntVector(Index, LastDim);
    DimIntVector(Index, Ubound);
    LastDim := Ubound;

    for I := Lbound to Ubound do
      begin
        Pvt := 0.0;
        for J := Lbound to Ubound do
          if Abs(A^[I]^[J]) > Pvt then
            Pvt := Abs(A^[I]^[J]);
        if Pvt < MACHEP then
          begin
            DelVector(V, Ubound);
            LU_Decomp := MAT_SINGUL;
            Exit;
          end;
        V^[I] := 1.0 / Pvt;
      end;
    for J := Lbound to Ubound do
      begin
        for I := Lbound to Pred(J) do
          begin
            Sum := A^[I]^[J];
            for K := Lbound to Pred(I) do
              Sum := Sum - A^[I]^[K] * A^[K]^[J];
            A^[I]^[J] := Sum;
          end;
        Imax := 0;
        Pvt := 0.0;
        for I := J to Ubound do
          begin
            Sum := A^[I]^[J];
            for K := Lbound to Pred(J) do
              Sum := Sum - A^[I]^[K] * A^[K]^[J];
            A^[I]^[J] := Sum;
            T := V^[I] * Abs(Sum);
            if T > Pvt then
              begin
                Pvt := T;
                Imax := I;
              end;
          end;
        if J <> Imax then
          begin
            SwapRows(Imax, J, A, Lbound, Ubound);
            V^[Imax] := V^[J];
          end;
        Index^[J] := Imax;
        if A^[J]^[J] = 0.0 then
          A^[J]^[J] := TINY;
        if J <> Ubound then
          begin
            T := 1.0 / A^[J]^[J];
            for I := Succ(J) to Ubound do
              A^[I]^[J] := A^[I]^[J] * T;
          end;
      end;
    DelVector(V, Ubound);
    LU_Decomp := MAT_OK;
  end;

  procedure LU_Solve(A : PMatrix; B : PVector; Lbound, Ubound : Integer;
                     X : PVector);
  var
    I, Ip, J, K : Integer;
    Sum : Float;
  begin
    K := Pred(Lbound);
    CopyVector(X, B, Lbound, Ubound);
    for I := Lbound to Ubound do
      begin
        Ip := Index^[I];
        Sum := X^[Ip];
        X^[Ip] := X^[I];
        if K >= Lbound then
          for J := K to Pred(I) do
            Sum := Sum - A^[I]^[J] * X^[J]
        else if Sum <> 0.0 then
          K := I;
        X^[I] := Sum;
      end;
    for I := Ubound downto Lbound do
      begin
        Sum := X^[I];
        if I < Ubound then
          for J := Succ(I) to Ubound do
            Sum := Sum - A^[I]^[J] * X^[J];
        X^[I] := Sum / A^[I]^[I];
      end;
  end;

  function SV_Decomp(A : PMatrix;
                     Lbound, Ubound1, Ubound2 : Integer;
                     S : PVector;
                     V : PMatrix) : Integer;
{ ----------------------------------------------------------------------
  This function is a translation of the EISPACK subroutine SVD

  This function determines the singular value decomposition
  A = U.S.V' of a real M by N rectangular matrix. Householder
  bidiagonalization and a variant of the QR algorithm are used.
  ----------------------------------------------------------------------
  This is a crude translation. Many of the original goto's
  have been kept!
  ---------------------------------------------------------------------- }

  var
    I, J, K, L, I1, K1, L1, Mn, Its : Integer;
    C, F, G, H, T, X, Y, Z, Tst1, Tst2, Scale : Float;
    R : PVector;

  label
    190, 210, 270, 290, 360, 390, 410, 430, 460,
    475, 490, 510, 520, 540, 565, 580, 650, 700;

  begin
    DimVector(R, Ubound2);
    Scale := 0.0;
    G := 0.0;
    X := 0.0;

    { Householder reduction to bidiagonal form }
    for I := Lbound to Ubound2 do
      begin
        L := I + 1;
        R^[I] := Scale * G;
        G := 0.0;
        T := 0.0;
        Scale := 0.0;
        if I > Ubound1 then goto 210;

        for K := I to Ubound1 do
          Scale := Scale + Abs(A^[K]^[I]);

        if Scale = 0.0 then goto 210;

        for K := I to Ubound1 do
          begin
            A^[K]^[I] := A^[K]^[I] / Scale;
            T := T + Sqr(A^[K]^[I]);
          end;

        F := A^[I]^[I];
        G := - DSgn(Sqrt(T), F);
        H := F * G - T;
        A^[I]^[I] := F - G;
        if I = Ubound2 then goto 190;

        for J := L to Ubound2 do
          begin
            T := 0.0;
            for K := I to Ubound1 do
              T := T + A^[K]^[I] * A^[K]^[J];
            F := T / H;
            for K := I to Ubound1 do
              A^[K]^[J] := A^[K]^[J] + F * A^[K]^[I];
          end;

190:    for K := I to Ubound1 do
          A^[K]^[I] := Scale * A^[K]^[I];

210:    S^[I] := Scale * G;
        G := 0.0;
        T := 0.0;
        Scale := 0.0;
        if (I > Ubound1) or (I = Ubound2) then goto 290;

        for K := L to Ubound2 do
          Scale := Scale + Abs(A^[I]^[K]);

        if Scale = 0.0 then goto 290;

        for K := L to Ubound2 do
          begin
            A^[I]^[K] := A^[I]^[K] / Scale;
            T := T + Sqr(A^[I]^[K]);
          end;

        F := A^[I]^[L];
        G := - DSgn(Sqrt(T), F);
        H := F * G - T;
        A^[I]^[L] := F - G;

        for K := L to Ubound2 do
          R^[K] := A^[I]^[K] / H;

        if I = Ubound1 then goto 270;

        for J := L to Ubound1 do
          begin
            T := 0.0;
            for K := L to Ubound2 do
              T := T + A^[J]^[K] * A^[I]^[K];
            for K := L to Ubound2 do
              A^[J]^[K] := A^[J]^[K] + T * R^[K];
          end;

270:    for K := L to Ubound2 do
          A^[I]^[K] := Scale * A^[I]^[K];

290:    X := FMax(X, Abs(S^[I]) + Abs(R^[I]));
      end;

    { Accumulation of right-hand transformations }
    for I := Ubound2 downto Lbound do
      begin
        if I = Ubound2 then goto 390;
        if G = 0.0 then goto 360;

        for J := L to Ubound2 do
          { Double division avoids possible underflow }
          V^[J]^[I] := (A^[I]^[J] / A^[I]^[L]) / G;

        for J := L to Ubound2 do
          begin
            T := 0.0;
            for K := L to Ubound2 do
              T := T + A^[I]^[K] * V^[K]^[J];
            for K := L to Ubound2 do
              V^[K]^[J] := V^[K]^[J] + T * V^[K]^[I];
          end;

360:    for J := L to Ubound2 do
          begin
            V^[I]^[J] := 0.0;
            V^[J]^[I] := 0.0;
          end;

390:    V^[I]^[I] := 1.0;
        G := R^[I];
        L := I;
      end;


410:{ Accumulation of left-hand transformations }
    Mn := IMin(Ubound1, Ubound2);

    for I := Mn downto Lbound do
      begin
        L := I + 1;
        G := S^[I];
        if I = Ubound2 then goto 430;

        for J := L to Ubound2 do
          A^[I]^[J] := 0.0;

430:    if G = 0.0 then goto 475;
        if I = Mn then goto 460;

        for J := L to Ubound2 do
          begin
            T := 0.0;

            for K := L to Ubound1 do
              T := T + A^[K]^[I] * A^[K]^[J];

            { Double division avoids possible underflow }
            F := (T / A^[I]^[I]) / G;

            for K := I to Ubound1 do
              A^[K]^[J] := A^[K]^[J] + F * A^[K]^[I];
          end;

460:    for J := I to Ubound1 do
          A^[J]^[I] := A^[J]^[I] / G;

        goto 490;

475:    for J := I to Ubound1 do
          A^[J]^[I] := 0.0;

490:    A^[I]^[I] := A^[I]^[I] + 1.0;
      end;

510:{ Diagonalization of the bidiagonal form }
    Tst1 := X;
    for K := Ubound2 downto Lbound do
      begin
        K1 := K - 1;
        Its := 0;

520:    { Test for splitting }
        for L := K downto Lbound do
          begin
            L1 := L - 1;
            Tst2 := Tst1 + Abs(R^[L]);
            if Tst2 = Tst1 then goto 565;
            { R^[Lbound] is always zero, so there is no exit
                through the bottom of the loop  }
            Tst2 := Tst1 + Abs(S^[L1]);
            if Tst2 = Tst1 then goto 540;
          end;

540:    { Cancellation of R^[L] if L greater than 1 }
        C := 0.0;
        T := 1.0;

        for I := L to K do
          begin
            F := T * R^[I];
            R^[I] := C * R^[I];
            Tst2 := Tst1 + Abs(F);
            if Tst2 = Tst1 then goto 565;
            G := S^[I];
            H := Pythag(F, G);
            S^[I] := H;
            C := G / H;
            T := - F / H;

            for J := Lbound to Ubound1 do
              begin
                Y := A^[J]^[L1];
                Z := A^[J]^[I];
                A^[J]^[L1] := Y * C + Z * T;
                A^[J]^[I] := - Y * T + Z * C;
              end;
           end;

565:    { Test for convergence }
        Z := S^[K];
        if L = K then goto 650;

        if Its = 30 then
          begin
            SV_Decomp := - K;
            Exit;
          end;

        { Shift from bottom 2 by 2 minor }
        Its := Its + 1;
        X := S^[L];
        Y := S^[K1];
        G := R^[K1];
        H := R^[K];
        F := 0.5 * (((G + Z) / H) * ((G - Z) / Y) + Y / H - H / Y);
        G := Pythag(F, 1.0);
        F := X - (Z / X) * Z + (H / X) * (Y / (F + DSgn(G, F)) - H);

        { Next QR transformation }
        C := 1.0;
        T := 1.0;

        for I1 := L to K1 do
          begin
            I := I1 + 1;
            G := R^[I];
            Y := S^[I];
            H := T * G;
            G := C * G;
            Z := Pythag(F, H);
            R^[I1] := Z;
            C := F / Z;
            T := H / Z;
            F := X * C + G * T;
            G := - X * T + G * C;
            H := Y * T;
            Y := Y * C;

            for J := Lbound to Ubound2 do
              begin
                X := V^[J]^[I1];
                Z := V^[J]^[I];
                V^[J]^[I1] := X * C + Z * T;
                V^[J]^[I] := - X * T + Z * C;
              end;

            Z := Pythag(F, H);
            S^[I1] := Z;
            { Rotation can be arbitrary if Z is zero }
            if Z = 0.0 then goto 580;
            C := F / Z;
            T := H / Z;
580:        F := C * G + T * Y;
            X := - T * G + C * Y;

            for J := Lbound to Ubound1 do
              begin
                Y := A^[J]^[I1];
                Z := A^[J]^[I];
                A^[J]^[I1] := Y * C + Z * T;
                A^[J]^[I] := - Y * T + Z * C;
              end;
           end;

        R^[L] := 0.0;
        R^[K] := F;
        S^[K] := X;
        goto 520;

650:    { Convergence }
        if Z >= 0.0 then goto 700;

        { S^[K] is made non-negative }
        S^[K] := - Z;

        for J := Lbound to Ubound2 do
          V^[J]^[K] := - V^[J]^[K];
700:  end;

    DelVector(R, Ubound2);
    SV_Decomp := 0;
  end;

  procedure SV_SetZero(S : PVector; Lbound, Ubound : Integer; Tol : Float);
  var
    Threshold : Float;
    I : Integer;
  begin
    Threshold := Tol * Max(S, Lbound, Ubound);
    for I := Lbound to Ubound do
      if S^[I] < Threshold then S^[I] := 0.0;
  end;

  procedure SV_Solve(U : PMatrix; S : PVector; V : PMatrix; B : PVector;
                     Lbound, Ubound1, Ubound2 : Integer;
                     X : PVector);
  var
    I, J, JJ : Integer;
    Sum : Float;
    Tmp : PVector;
  begin
    DimVector(Tmp, Ubound2);
    for J := Lbound to Ubound2 do
      begin
        Sum := 0.0;
        if S^[J] > 0.0 then
          begin
            for I := Lbound to Ubound1 do
              Sum := Sum + U^[I]^[J] * B^[I];
            Sum := Sum / S^[J];
          end;
        Tmp^[J] := Sum;
      end;
    for J := Lbound to Ubound2 do
      begin
        Sum := 0.0;
        for JJ := Lbound to Ubound2 do
          Sum := Sum + V^[J]^[JJ] * Tmp^[JJ];
        X^[J] := Sum;
      end;
    DelVector(Tmp, Ubound2);
  end;

  procedure SV_Approx(U : PMatrix; S : PVector; V : PMatrix;
                      Lbound, Ubound1, Ubound2 : Integer; A : PMatrix);
  var
    I, J, K : Integer;
  begin
    for I := Lbound to Ubound1 do
      for J := Lbound to Ubound2 do
        begin
          A^[I]^[J] := 0.0;
          for K := Lbound to Ubound2 do
            if S^[K] > 0.0 then
              A^[I]^[J] := A^[I]^[J] + U^[I]^[K] * V^[J]^[K];
        end;
  end;

  function QR_Decomp(A : PMatrix; Lbound, Ubound1, Ubound2 : Integer;
                     R : PMatrix) : Integer;
  var
    I, J, K : Integer;
    Sum : Float;
  begin
    for K := Lbound to Ubound2 do
      begin
        { Compute the "k"th diagonal entry in R }
        Sum := 0.0;
        for I := Lbound to Ubound1 do
          Sum := Sum + Sqr(A^[I]^[K]);

        if Sum = 0.0 then
          begin
            QR_Decomp := MAT_SINGUL;
            Exit;
          end;

        R^[K]^[K] := Sqrt(Sum);

        { Divide the entries in the "k"th column of A by the computed "k"th }
        { diagonal element of R.  this begins the process of overwriting A  }
        { with Q . . .                                                      }
        for I := Lbound to Ubound1 do
          A^[I]^[K] := A^[I]^[K] / R^[K]^[K];

        for J := (K + 1) to Ubound2 do
          begin
            { Complete the remainder of the row entries in R }
            Sum := 0.0;
            for I := Lbound to Ubound1 do
              Sum := Sum + A^[I]^[K] * A^[I]^[J];
            R^[K]^[J] := Sum;

            { Update the column entries of the Q/A matrix }
            for I := Lbound to Ubound1 do
              A^[I]^[J] := A^[I]^[J] - A^[I]^[K] * R^[K]^[J];
          end;
      end;

    QR_Decomp := MAT_OK;
  end;

  procedure QR_Solve(Q, R : PMatrix; B : PVector;
                     Lbound, Ubound1, Ubound2 : Integer;
                     X : PVector);
  var
    I, J : Integer;
    Sum : Float;
  begin
    { Form Q'B and store the result in X }
    for J := Lbound to Ubound2 do
      begin
        X^[J] := 0.0;
        for I := Lbound to Ubound1 do
          X^[J] := X^[J] + Q^[I]^[J] * B^[I];
      end;

    { Update X with the solution vector }
    X^[Ubound2] := X^[Ubound2] / R^[Ubound2]^[Ubound2];
    for I := (Ubound2 - 1) downto Lbound do
      begin
        Sum := 0.0;
        for J := (I + 1) to Ubound2 do
          Sum := Sum + R^[I]^[J] * X^[J];
        X^[I] := (X^[I] - Sum) / R^[I]^[I];
      end;
  end;

begin
  { Initialize global variables for LU }
  LastDim := 1;
  Index := nil;
end.
