{ **********************************************************************
  *                         Unit PLOTVAR.PAS                           *
  *                            Version 1.1                             *
  *                   (c) J. Debord, February 2002                     *
  **********************************************************************
            Graphic constants, types, variables and procedures
                     used by units PLOT and TEXPLOT
  ********************************************************************** }

unit plotvar;

interface

uses
  fmath, matrices;

const
  MAXSYMBOL = 9;        { Max. number of graphic symbols }
  EPS       = 1.0E-10;  { Lower limit for an axis label }

type
  TScale  = (LIN_SCALE,        { Scale }
             LOG_SCALE);

  TGrid   = (NO_GRID,          { Grid }
             HORIZ_GRID,
             VERTIC_GRID,
             BOTH_GRID);

  TAxis   = record             { Coordinate axis }
              Scale : TScale;
              Min   : Float;
              Max   : Float;
              Step  : Float;
            end;

  TTitle  = record             { Title for main graph or axis }
              Text       : String[70];
              Font       : Integer;
              CharWidth  : Integer;
              CharHeight : Integer;
            end;

  TLegend = record             { Legends of plotted curves }
              Text       : array[1..MAXSYMBOL] of String[40];
              Font       : Integer;
              CharWidth  : Integer;
              CharHeight : Integer;
              SymbolSize : Integer;
            end;

var
  Xwin1, Ywin1, Xwin2, Ywin2 : Integer; { Window limits in % }

  GraphBorder : Boolean;  { Plot graph border }
  XAxis       : TAxis;    { Horizontal axis }
  YAxis       : TAxis;    { Vertical axis }
  Grid        : TGrid;    { Grid }
  GraphTitle  : TTitle;   { Title of graph }
  XTitle      : TTitle;   { Title of X axis }
  YTitle      : TTitle;   { Title of Y axis }
  Legend      : TLegend;  { Legends of curves }

procedure Interval(X1, X2 : Float; MinDiv, MaxDiv : Integer;
                   var Min, Max, Step : Float);
{ ----------------------------------------------------------------------
  Determines an interval [Min, Max] including the values from X1 to X2,
  and a subdivision Step of this interval
  ----------------------------------------------------------------------
  Input parameters  : X1, X2 = min. & max. values to be included
                      MinDiv = minimum nb of subdivisions
                      MaxDiv = maximum nb of subdivisions
  ----------------------------------------------------------------------
  Output parameters : Min, Max, Step
  ---------------------------------------------------------------------- }

procedure AutoScale(Z : PVector; Lbound, Ubound : Integer;
                    var Axis : TAxis);
{ ----------------------------------------------------------------------
  Determines the scale of an axis
  ----------------------------------------------------------------------
  Input parameters  : Z      = array of values to be plotted
                      Lbound,
                      Ubound = indices of first and last elements of Z
  ----------------------------------------------------------------------
  Output parameters : Axis
  ---------------------------------------------------------------------- }

implementation

  procedure Interval(X1, X2 : Float; MinDiv, MaxDiv : Integer;
                     var Min, Max, Step : Float);
  var
    H, R, K : Float;
  begin
    if X1 >= X2 then Exit;
    H := X2 - X1;
    R := Int(Log10(H));
    if H < 1.0 then R := R - 1.0;
    Step := Exp10(R);

    repeat
      K := Int(H / Step);
      if K < MinDiv then Step := 0.5 * Step;
      if K > MaxDiv then Step := 2.0 * Step;
    until (K >= MinDiv) and (K <= MaxDiv);

    Min := Step * Int(X1 / Step);
    Max := Step * Int(X2 / Step);
    while Min > X1 do Min := Min - Step;
    while Max < X2 do Max := Max + Step;
  end;

  procedure AutoScale(Z : PVector; Lbound, Ubound : Integer;
                      var Axis : TAxis);
  var
    I                  : Integer;
    Zmin, Zmax, Z1, Z2 : Float;
  begin
    if Axis.Scale = LIN_SCALE then
      Interval(Min(Z, Lbound, Ubound), Max(Z, Lbound, Ubound),
               2, 6, Axis.Min, Axis.Max, Axis.Step)
    else
      begin
        Zmin := MAXNUM; Zmax := 0.0;
        for I := Lbound to Ubound do
          if Z^[I] > 0.0 then
            if Z^[I] < Zmin then
              Zmin := Z^[I]
            else if Z^[I] > Zmax then
              Zmax := Z^[I];
        Z1 := Int(Log10(Zmin));
        Z2 := Int(Log10(Zmax));
        if Zmin < 1.0 then Z1 := Z1 - 1.0;
        if Zmax > 1.0 then Z2 := Z2 + 1.0;
        Axis.Min := Z1;
        Axis.Max := Z2;
        Axis.Step := 1.0;
      end;
  end;

var
  I : Integer;

begin
  Xwin1 := 15;
  Ywin1 := 15;
  Xwin2 := 85;
  Ywin2 := 85;

  GraphBorder := True;

  with XAxis do
    begin
      Scale := LIN_SCALE;
      Min   := 0.0;
      Max   := 1.0;
      Step  := 0.2;
    end;

  with YAxis do
    begin
      Scale := LIN_SCALE;
      Min   := 0.0;
      Max   := 1.0;
      Step  := 0.2;
    end;

  Grid := NO_GRID;

  with GraphTitle do
    begin
      Text       := '';
      Font       := 2;
      CharWidth  := 300;
      CharHeight := 350;
    end;

  with XTitle do
    begin
      Text       := 'X';
      Font       := 2;
      CharWidth  := 200;
      CharHeight := 250;
    end;

  with YTitle do
    begin
      Text       := 'Y';
      Font       := 2;
      CharWidth  := 200;
      CharHeight := 250;
    end;

  with Legend do
    begin
      Font       := 2;
      CharWidth  := 50;
      CharHeight := 50;
      SymbolSize := 3;

      for I := 1 to MAXSYMBOL do
        Text[I] := Chr(32 + I);
    end;

end.